@@ -19,14 +19,16 @@
 #include <media/medium.hxx>
 #include <types.hxx>
 
-#define MIN(a,b) (a<b?a:b)
-
 using std::auto_ptr;
 using std::string;
 using std::exception;
 using std::list;
 namespace off = offlinefs;
 
+// Helper functions
+
+#define MIN(a,b) (a<b?a:b)
+
 inline SContext FS::userctx(){
    return scache.get(fuse_get_context()->uid,fuse_get_context()->gid);
 }
@@ -66,6 +68,16 @@ int FS::errcode(exception& e){
    }
 }
 
+inline void FS::check_sticky_bit(SContext& sctx, Directory* parent, Node* n){
+   if((parent->getattr<off::mode_t>("offlinefs.mode") & S_ISVTX) &&
+      (sctx.uid != 0)  &&
+      (sctx.uid != parent->getattr<off::uid_t>("offlinefs.uid")) &&
+      (sctx.uid != n->getattr<off::uid_t>("offlinefs.uid")))
+      throw Node::EAccess();
+}
+
+// Filesystem operations
+
 FS::FS(string dbroot,string defmedium):dbs(dbroot),defmedium(defmedium) {
    memset(openFiles,0,sizeof(openFiles));
    dbs.open();
@@ -99,10 +111,11 @@ int FS::getattr(const char* path, struct stat* st){
       st->st_atime=n->getattr<off::time_t>("offlinefs.atime");
       st->st_mtime=n->getattr<off::time_t>("offlinefs.mtime");
       st->st_ctime=n->getattr<off::time_t>("offlinefs.ctime");
+      st->st_blocks=st->st_size/512;
    }catch(exception& e){
       return errcode(e);
    }
-  return 0;
+   return 0;
 }
 
 int FS::readdir(const char* path, void* buf, fuse_fill_dir_t filler, off_t offset, fuse_file_info* fi){ 
@@ -115,13 +128,16 @@ int FS::readdir(const char* path, void* buf, fuse_fill_dir_t filler, off_t offse
 
       n->setattr<off::time_t>("offlinefs.atime",time(NULL));
 
-      list<string> l=n->getchildren();
+      list<string> l;
+      n->getchildren(l);
+
       for(list<string>::iterator it=l.begin();it!=l.end();it++)
 	 filler(buf,it->c_str(),NULL,0);
+
    }catch(exception& e){
       return errcode(e);
    }
-  return 0;
+   return 0;
 }
 
 int FS::readlink(const char* path, char* buf, size_t bufsiz){
@@ -155,6 +171,11 @@ int FS::mknod(const char* path , mode_t mode, dev_t dev){
       n->setattr<off::mode_t>("offlinefs.mode",mode);
       if(S_ISCHR(mode)||S_ISBLK(mode))
 	 n->setattr<off::dev_t>("offlinefs.dev",dev);
+
+      // Update the parent time attributes
+      p.parent->setattr<off::time_t>("offlinefs.ctime",time(NULL));
+      p.parent->setattr<off::time_t>("offlinefs.mtime",time(NULL));
+
    }catch(exception& e){
       return errcode(e);
    }
@@ -173,6 +194,11 @@ int FS::mkdir(const char* path, mode_t mode){
       n->setattr<off::uid_t>("offlinefs.uid",fuse_get_context()->uid);
       n->setattr<off::gid_t>("offlinefs.gid",fuse_get_context()->gid);
       n->setattr<off::mode_t>("offlinefs.mode",S_IFDIR|mode);
+
+      // Update the parent time attributes
+      p.parent->setattr<off::time_t>("offlinefs.ctime",time(NULL));
+      p.parent->setattr<off::time_t>("offlinefs.mtime",time(NULL));
+
    }catch(exception& e){
       return errcode(e);
    }
@@ -187,10 +213,22 @@ int FS::unlink(const char* path){
 
       p.parent->access(sctx,W_OK|X_OK);
 
-      p.parent->getchild(p.leaf)->setattr<off::time_t>("offlinefs.ctime",time(NULL));
+      auto_ptr<Node> n = p.parent->getchild(p.leaf);
 
+      check_sticky_bit(sctx, p.parent.get(), n.get());
+
+      // Don't unlink directories
+      if(S_ISDIR(n->getattr<off::mode_t>("offlinefs.mode")))
+	 return -EISDIR;
+
+      // Unlink it
+      pcache.invalidate(txns, path);
+      n->setattr<off::time_t>("offlinefs.ctime",time(NULL));
       p.parent->delchild(p.leaf);
-      pcache.invalidate(path);
+
+      // Update the parent time attributes
+      p.parent->setattr<off::time_t>("offlinefs.ctime",time(NULL));
+      p.parent->setattr<off::time_t>("offlinefs.mtime",time(NULL));
 
    }catch(exception& e){
       return errcode(e);
@@ -207,11 +245,24 @@ int FS::rmdir(const char* path){
       p.parent->access(sctx,X_OK|W_OK);
 
       auto_ptr<Directory> n=Node::cast<Directory>(p.parent->getchild(p.leaf));
-      if(n->getchildren().size()>2)
+
+      check_sticky_bit(sctx, p.parent.get(), n.get());
+
+      // It should only contain "." and ".."
+      if(n->countchildren() > 2)
 	 return -ENOTEMPTY;
 
+      n->delchild(".");
+      n->delchild("..");
+
+      // Unlink it
       p.parent->delchild(p.leaf);
-      pcache.invalidate(path);
+      pcache.invalidate(txns,path);
+
+      // Update the parent time attributes
+      p.parent->setattr<off::time_t>("offlinefs.ctime",time(NULL));
+      p.parent->setattr<off::time_t>("offlinefs.mtime",time(NULL));
+
    }catch(exception& e){
       return errcode(e);
    }
@@ -229,6 +280,11 @@ int FS::symlink(const char* oldpath, const char* newpath){
       sl->setattr<off::uid_t>("offlinefs.uid",fuse_get_context()->uid);
       sl->setattr<off::gid_t>("offlinefs.gid",fuse_get_context()->gid);
       sl->setattrv("offlinefs.symlink",Buffer(oldpath));
+
+      // Update the parent time attributes
+      p.parent->setattr<off::time_t>("offlinefs.ctime",time(NULL));
+      p.parent->setattr<off::time_t>("offlinefs.mtime",time(NULL));
+
    }catch(exception& e){
       return errcode(e);
    }
@@ -236,11 +292,86 @@ int FS::symlink(const char* oldpath, const char* newpath){
 }
  
 int FS::rename(const char* oldpath, const char* newpath){
-   int err;
-   if((err=link(oldpath,newpath)))
-      return err;
-   if((err=unlink(oldpath)))
-      return err;
+   try{
+      FsTxn txns(dbs);
+      SContext sctx=userctx();
+      Path<Node> srcpath(txns,sctx,pcache,oldpath);
+      Path<Node> dstpath(txns,sctx,pcache,newpath);
+      auto_ptr<Node> src,dst;
+
+      src = srcpath.parent->getchild(srcpath.leaf);
+      try{
+	 dst = dstpath.parent->getchild(dstpath.leaf);
+      }catch(Node::ENotFound& e){}
+
+      // Check if both paths refer to the same node
+      if(dst.get() && string(oldpath) == string(newpath))
+	 return 0;
+
+      srcpath.parent->access(sctx,X_OK|W_OK);
+      dstpath.parent->access(sctx,X_OK|W_OK);
+
+      check_sticky_bit(sctx, srcpath.parent.get(), src.get());
+
+      if(dst.get()){
+	 check_sticky_bit(sctx, dstpath.parent.get(), dst.get());
+
+	 // Ensure that both nodes are directories, if any
+	 if(S_ISDIR(src->getattr<off::mode_t>("offlinefs.mode"))){
+	    if(!S_ISDIR(dst->getattr<off::mode_t>("offlinefs.mode")))
+	       return -ENOTDIR;
+
+	    // Check if dst is a subdirectory of src
+	    for(auto_ptr<Directory> n=Node::cast<Directory>(Node::getnode(dstpath.parent.get()));
+		n->getid(); n = Node::cast<Directory>(n->getchild(".."))){
+	       if(n->getid() == src->getid())
+		  return -EINVAL;
+	    }
+
+	    // Ensure that the destination node is empty
+	    auto_ptr<Directory> d = Node::cast<Directory>(Node::getnode(dst.get()));
+
+	    if(d->countchildren() > 2)
+	       return -ENOTEMPTY;
+
+	    d->delchild(".");
+	    d->delchild("..");
+
+	 }else{
+	    if(S_ISDIR(dst->getattr<off::mode_t>("offlinefs.mode")))
+	       return -EISDIR;
+	 }
+
+	 // Remove the destination node
+	 pcache.invalidate(txns, newpath);
+	 dst->setattr<off::time_t>("offlinefs.ctime",time(NULL));
+	 dstpath.parent->delchild(dstpath.leaf);
+      }
+
+      if(S_ISDIR(src->getattr<off::mode_t>("offlinefs.mode"))){
+	 Directory* d = (Directory*)src.get();
+
+	 // Update the parent of src
+	 d->delchild("..");
+	 d->addchild("..",*dstpath.parent);
+      }
+
+      // Relink src at the destination path
+      pcache.invalidate(txns, oldpath);
+      src->setattr<off::time_t>("offlinefs.ctime",time(NULL));
+
+      dstpath.parent->addchild(dstpath.leaf, *src);
+      srcpath.parent->delchild(srcpath.leaf);
+
+      // Update the ctime/mtime attributes of the modified directories
+      srcpath.parent->setattr<off::time_t>("offlinefs.ctime",time(NULL));
+      srcpath.parent->setattr<off::time_t>("offlinefs.mtime",time(NULL));
+      dstpath.parent->setattr<off::time_t>("offlinefs.ctime",time(NULL));
+      dstpath.parent->setattr<off::time_t>("offlinefs.mtime",time(NULL));
+
+   }catch(exception& e){
+      return errcode(e);
+   }
    return 0;
 }
 
@@ -256,6 +387,11 @@ int FS::link(const char* oldpath, const char* newpath){
 
       n->setattr<off::time_t>("offlinefs.ctime",time(NULL));
       p.parent->addchild(p.leaf,*n);
+
+      // Update the parent time attributes
+      p.parent->setattr<off::time_t>("offlinefs.ctime",time(NULL));
+      p.parent->setattr<off::time_t>("offlinefs.mtime",time(NULL));
+
    }catch(exception& e){
       return errcode(e);
    }
@@ -268,7 +404,7 @@ int FS::chmod(const char* path, mode_t mode){
       SContext sctx=userctx();
       auto_ptr<Node> n=pcache.getnode(txns,sctx,path);
 
-      if(fuse_get_context()->uid!=0){
+      if(sctx.uid != 0){
 	 gid_t fgid=n->getattr<off::gid_t>("offlinefs.gid");
 	 uid_t fuid=n->getattr<off::uid_t>("offlinefs.uid");
 	 if(sctx.uid!=fuid)
@@ -279,7 +415,7 @@ int FS::chmod(const char* path, mode_t mode){
 
       n->setattr<off::time_t>("offlinefs.ctime",time(NULL));
       n->setattr<off::mode_t>("offlinefs.mode",(mode&(~S_IFMT))|(n->getattr<off::mode_t>("offlinefs.mode")&S_IFMT));
-      pcache.invalidateAccess(path);
+      pcache.invalidate(txns,path);
    }catch(exception& e){
       return errcode(e);
    }
@@ -294,19 +430,25 @@ int FS::chown(const char* path, uid_t owner, gid_t group){
 
       if(owner!=(uid_t)-1 && owner!=n->getattr<off::uid_t>("offlinefs.uid")){
 	 if(sctx.uid!=0)
-	    return -EACCES;
+	    return -EPERM;
 	 n->setattr<off::uid_t>("offlinefs.uid",owner);
       }
 
       if(group!=(gid_t)-1 && group!=n->getattr<off::gid_t>("offlinefs.gid")){
 	 if(sctx.uid!=0){
-	    if(!sctx.groups.count(group))
-	       return -EACCES;
-	    else{
-	       //Clear setuid/setgid bit
-	       mode_t m=n->getattr<off::mode_t>("offlinefs.mode")&(~S_ISUID);
-	       if(m&S_IXGRP)
+	    if(sctx.uid != n->getattr<off::uid_t>("offlinefs.uid") ||
+	       !sctx.groups.count(group))
+	       return -EPERM;
+
+	    //Clear setuid/setgid bit
+	    mode_t m=n->getattr<off::mode_t>("offlinefs.mode");
+
+	    if(m & S_IFREG){
+	       m&=~S_ISUID;
+
+	       if(m & S_IXGRP)
 		  m&=~S_ISGID;
+
 	       n->setattr<off::mode_t>("offlinefs.mode",m);
 	    }
 	 }
@@ -314,7 +456,7 @@ int FS::chown(const char* path, uid_t owner, gid_t group){
       }
 
       n->setattr<off::time_t>("offlinefs.ctime",time(NULL));
-      pcache.invalidateAccess(path);
+      pcache.invalidate(txns,path);
    }catch(exception& e){
       return errcode(e);
    }
@@ -384,12 +526,6 @@ int FS::read(const char* path, char* buf, size_t nbyte, off_t offset, struct fus
    if(fi->fh>=MAX_OPEN_FILES || !openFiles[fi->fh])
       return -EBADF;
 
-   if(nbyte>0){
-      FsTxn txns(dbs);
-      File f(txns,openFiles[fi->fh]->getfileid());
-      f.setattr<off::time_t>("offlinefs.atime",time(NULL));
-   }
-
    return openFiles[fi->fh]->read(buf,nbyte,offset);
 }
 
@@ -397,12 +533,6 @@ int FS::write(const char* path, const char* buf, size_t nbyte, off_t offset, str
    if(fi->fh>=MAX_OPEN_FILES || !openFiles[fi->fh])
       return -EBADF;
 
-   if(nbyte>0){
-      FsTxn txns(dbs);
-      File f(txns,openFiles[fi->fh]->getfileid());
-      f.setattr<off::time_t>("offlinefs.mtime",time(NULL));
-   }
-
    return openFiles[fi->fh]->write(buf,nbyte,offset);
 }
 
@@ -411,21 +541,23 @@ int FS::statfs(const char* path, struct statvfs* buf){
       FsTxn txns(dbs);
       Medium::Stats st;
 
-      list<uint32_t> ms = dbs.media.listregisters(txns.media);
+      list<uint32_t> ms;
+      dbs.media.getregisters(txns.media,ms);
+
       for(list<uint32_t>::iterator it=ms.begin();
 	  it!=ms.end(); it++){
 	 Medium::Stats st_=(Medium::getmedium(txns, *it))->getstats();
 
-	 st.blocks+=st_.blocks;
-	 st.freeblocks+=st_.freeblocks;
+	 st.total+=st_.total;
+	 st.free+=st_.free;
       }
 
       memset(buf,0,sizeof(struct statvfs));
       buf->f_bsize=4096;
       buf->f_frsize=4096;
-      buf->f_blocks=st.blocks;
-      buf->f_bfree=st.freeblocks;
-      buf->f_bavail=st.freeblocks;
+      buf->f_blocks=st.total/4096;
+      buf->f_bfree=st.free/4096;
+      buf->f_bavail=st.free/4096;
       buf->f_namemax=0;
 
       return 0;
@@ -464,7 +596,8 @@ int FS::setxattr(const char* path, const char* name, const char* value, size_t s
       SContext sctx=userctx();
       auto_ptr<Node> n=pcache.getnode(txns,sctx,path);
 
-      if(sctx.uid!=0 && sctx.uid!=getuid() && (string(name).find("offlinefs.")==0 || sctx.uid!=n->getattr<off::uid_t>("offlinefs.uid")) )
+      if(sctx.uid!=0 && (string(name).find("offlinefs.")==0 ||
+			 sctx.uid!=n->getattr<off::uid_t>("offlinefs.uid")))
 	 return -EACCES;
 
       n->setattr<off::time_t>("offlinefs.ctime",time(NULL));
@@ -484,8 +617,9 @@ int FS::setxattr(const char* path, const char* name, const char* value, size_t s
 
       n->setattrv(name,Buffer(value,size));
 
-      if(string(name)=="offlinefs.uid" || string(name)=="offlinefs.gid" || string(name)=="offlinefs.mode")
-	 pcache.invalidateAccess(path);
+      if(string(name)=="offlinefs.uid" || string(name)=="offlinefs.gid" ||
+	 string(name)=="offlinefs.mode")
+	 pcache.invalidate(txns,path);
 
    }catch(Node::EAttrNotFound& e){
       return -ENOATTR;
@@ -518,7 +652,8 @@ int FS::listxattr(const char* path , char* list, size_t size){
       SContext sctx=userctx();
       auto_ptr<Node> n=pcache.getnode(txns,sctx,path);
 
-      std::list<string> l=n->getattrs();
+      std::list<string> l;
+      n->getattrs(l);
 
       size_t count=0;
       for(std::list<string>::iterator it=l.begin();it!=l.end();it++){
@@ -542,7 +677,8 @@ int FS::removexattr(const char* path, const char* name){
       SContext sctx=userctx();
       auto_ptr<Node> n=pcache.getnode(txns,sctx,path);
 
-      if(sctx.uid!=0 && sctx.uid!=getuid() && (string(name).find("offlinefs.")==0 || sctx.uid!=n->getattr<off::uid_t>("offlinefs.uid")) )
+      if(sctx.uid!=0 && (string(name).find("offlinefs.")==0 ||
+			 sctx.uid!=n->getattr<off::uid_t>("offlinefs.uid")))
 	 return -EACCES;
 
       n->setattr<off::time_t>("offlinefs.ctime",time(NULL));
@@ -563,7 +699,7 @@ int FS::utimens(const char* path, const struct timespec tv[2]){
       auto_ptr<Node> n=pcache.getnode(txns,sctx,path);
 
       if(tv){
-	 if(sctx.uid!=0 && sctx.uid!=getuid() && sctx.uid!=n->getattr<off::uid_t>("offlinefs.uid"))
+	 if(sctx.uid!=0 && sctx.uid!=n->getattr<off::uid_t>("offlinefs.uid"))
 	    return -EACCES;
 
 	 n->setattr<off::time_t>("offlinefs.atime",tv[0].tv_sec);
@@ -594,3 +730,17 @@ int FS::access(const char* path, int mode){
    }
    return 0;
 }
+
+int FS::opendir(const char* path, struct fuse_file_info* fi){
+   try{
+      FsTxn txns(dbs);
+      SContext sctx=userctx();
+      auto_ptr<Directory> n=Node::cast<Directory>(pcache.getnode(txns,sctx,path));
+
+      n->access(sctx,R_OK|X_OK);
+
+   }catch(exception& e){
+      return errcode(e);
+   }
+   return 0;
+}
